""" Contains Materialflow and MaterialFlowCallbacks class """

# standard libraries
import uuid
import networkx as nx
import matplotlib.pyplot as plt

# local sources
from lotlan_scheduler.api.event import Event
from lotlan_scheduler.api.transportorder import TransportOrder

from lotlan_scheduler.logger.sqlite_logger import SQLiteLogger

from lotlan_scheduler.petri_net.logic import PetriNetLogic

from lotlan_scheduler.petri_net.generator import PetriNetGenerator

# globals defines
from lotlan_scheduler.defines import PetriNetConstants, LogicConstants

class MaterialFlowCallbacks(object):
    """
        Contains lists of registered callback functions
        for different states in the scheduling process
    """
    def __init__(self):
        self.triggered_by_cb = []
        self.pickup_finished_cb = []
        self.delivery_finished_cb = []
        self.finished_by_cb = []
        self.next_to_cb = []
        self.task_finished_cb = []
        self.all_finished_cb = []

class MaterialFlow():
    """ Represents an abstract materialflow """

    def __init__(self, _uuid, lotlan_structure, lotlan_string, tasks_in_mf, test_flag=False):
        self._uuid = _uuid
        self.name = ""
        self._is_running = True
        self.materialflow_callbacks = MaterialFlowCallbacks()
        self.tasks_in_mf = tasks_in_mf
        self.lotlan_structure = lotlan_structure
        self.tasks = {}
        self.ids = {}
        self.triggered_by_events = {}
        self.finished_by_events = {}
        self.event_instances = {}
        self.not_done_parents = {}  # 0 if all parent tasks are done
        self.tasks_done = {}
        self.test_flag = test_flag
        self.parent_count = {}
        self.lotlan_string = lotlan_string
        self.petri_net_generator = PetriNetGenerator(tasks_in_mf,
                                                     self.event_instances,
                                                     test_flag=test_flag)
        self.logger = None
        self.call_graph = None
        self.startable_tasks = None
        self.cycle_in_call_graph = None
        self.petri_net_logic = None

    def is_running(self):
        return self._is_running

    def start(self):
        """
            Starts the materialflow scheduling
        """
        self.logger = SQLiteLogger()
        self.logger.insert_materialflow_in_sql(self._uuid, self.lotlan_string)

        self.initialize_tasks(self.tasks_in_mf)

        if self.tasks_in_mf:
            self.name = self.tasks_in_mf[0].name

        self.call_graph = self.create_call_graph(self.tasks_in_mf)

        cycles = list(nx.simple_cycles(self.call_graph))
        self.cycle_in_call_graph = len(cycles) > 0

        self.startable_tasks = self.find_startable_tasks(self.call_graph, self.tasks_in_mf)

        for instance in self.lotlan_structure.instances.values():
            if instance.template_name == "Event":
                self.event_instances[instance.name] = instance
        task_representations = self.petri_net_generator.generate_task_nets()
        self.petri_net_logic = PetriNetLogic(task_representations, self.test_flag)

        self.create_event_information_list()
        self.start_tasks(self.startable_tasks)

    def start_tasks(self, tasks):
        """
            Starts scheduling of the given tasks

            if a task has a triggeredBy statement it waits for incoming events
            otherwise the transport_order can be executed and so next_to is called
        """
        next_tos = []
        for task in tasks:
            uuid_ = self.ids[task.name]
            transport_order = task.transport_order
            pickup = transport_order.pickup_tos.location
            delivery = transport_order.delivery_tos.location

            transport_order.state = TransportOrder.TransportOrderState.TASK_STARTED
            state = transport_order.state
            self.logger.insert_transport_order(self._uuid, uuid_, state, pickup, delivery)

            if self.triggered_by_events[task.name]:
                tb_events_of_task = self.triggered_by_events[task.name]
                self.petri_net_logic.set_awaited_events(task, tb_events_of_task)

                self.wait_for_triggered_by(uuid_, self.triggered_by_events[task.name])

                transport_order.state = TransportOrder.TransportOrderState.TASK_WAIT_FOR_TRIGGERED_BY
                state = transport_order.state
                self.logger.insert_transport_order(self._uuid, uuid_, state, pickup, delivery)
            else:
                task_started_event = Event(PetriNetConstants.TASK_STARTED_PLACE, "", "Boolean",
                                           comparator="", value=True)
                self.petri_net_logic.set_awaited_events(task, [task_started_event])
                self.fire_event(uuid_, task_started_event)
                next_tos.append(task)

        self.next_to(next_tos)

    def create_call_graph(self, tasks):
        """
            Creates a graph where every node is a task
            and a directed edge represents an onDone
        """
        call_graph = nx.DiGraph()
        for task in tasks:
            call_graph.add_node(task.name)
            for child in task.on_done:
                call_graph.add_edge(task.name, child)
        return call_graph

    def save_call_graph_img(self, filename):
        """ Saves the generated call graph of the materialflow as image """
        nx.draw(self.call_graph, with_labels=True)
        plt.savefig(filename, dpi=300, bbox_inches="tight")

    def find_startable_tasks(self, graph, tasks):
        """
            Finds tasks that can be started:
            task with no incoming edges in graph
        """
        startable_tasks = []
        for task in tasks:
            incoming_edges = 0
            for u, v in graph.in_edges(task.name):
                # ignore self loops
                if u != v:
                    incoming_edges = incoming_edges + 1
                else:
                    self.not_done_parents[task.name] = 1
            self.not_done_parents[task.name] = self.not_done_parents[task.name] + incoming_edges
            self.parent_count[task.name] = self.not_done_parents[task.name]
            if incoming_edges == 0:
                startable_tasks.append(task)

        return startable_tasks

    def fire_event(self, uuid_, event):
        """ Fires event to petri net corresponding to task of uuid """
        task = self.tasks[str(uuid_)]
        self.petri_net_logic.fire_event(task, event, self.on_petri_net_response)

    def initialize_tasks(self, tasks):
        """ Adds information for api classes to tasks and init dicts """
        for i, task in enumerate(tasks):
            if self.test_flag:
                uuid_ = i
            else:
                uuid_ = uuid.uuid4()
            self.tasks[str(uuid_)] = task
            self.ids[task.name] = uuid_
            self.tasks_done[task.name] = False
            self.not_done_parents[task.name] = 0

            transport_order = task.transport_order
            transport_order.uuid = uuid_
            pickup_tos = transport_order.pickup_tos
            delivery_tos = transport_order.delivery_tos

            for instance in self.lotlan_structure.instances.values():
                if instance.template_name == "Location":
                    if pickup_tos.location.logical_name == instance.name:
                        pickup_tos.location.physical_name = instance.keyval["name"]
                        pickup_tos.location.location_type = instance.keyval["type"]
                    elif delivery_tos.location.logical_name == instance.name:
                        delivery_tos.location.physical_name = instance.keyval["name"]
                        delivery_tos.location.location_type = instance.keyval["type"]

    def create_event_information_list(self):
        """ Creates a list of events objects out of the event names """
        for task in self.tasks_in_mf:
            triggered_by = []
            for event_name in task.triggered_by_events:
                logical_name = self.event_instances[event_name].name
                physical_name = self.event_instances[event_name].keyval["name"]
                event_type = self.event_instances[event_name].keyval["type"]
                triggered_by.append(Event(logical_name, physical_name, event_type, None, None))
            self.triggered_by_events[task.name] = triggered_by

            finished_by = []
            for event_name in task.finished_by_events:
                logical_name = self.event_instances[event_name].name
                physical_name = self.event_instances[event_name].keyval["name"]
                event_type = self.event_instances[event_name].keyval["type"]
                finished_by.append(Event(logical_name, physical_name, event_type, None, None))
            self.finished_by_events[task.name] = finished_by

    def on_petri_net_response(self, msg, task):
        """
            Handles incoming messages from the petri net logic and
            calls corresponding methods
        """
        if msg == LogicConstants.TRIGGERED_BY_PASSED_MSG:
            self.next_to([task])
        elif msg == LogicConstants.TOS_TB_PASSED_MSG:
            self.on_tos_tb_passed(task)
        elif msg == LogicConstants.TOS_WAIT_FOR_ACTION:
            self.on_tos_wait_for_action(task)
        elif msg == LogicConstants.TOS_FINISHED_MSG:
            self.on_tos_finished(task)
        elif msg == LogicConstants.TO_DONE_MSG:
            self.on_to_done(task)
        elif msg == LogicConstants.TASK_FINISHED_MSG:
            self.on_task_finished(task)

    def next_to(self, task_info):
        """ Notifies listeners about the next transport orders and set petri net state """
        if task_info:
            transport_orders = {}
            for task in task_info:
                uid = self.ids[task.name]

                transport_order = task.transport_order
                task.transport_order.state = TransportOrder.TransportOrderState.PICKUP_STARTED
                transport_orders[uid] = transport_order

                to_done_event = Event("to_done", "", "Boolean",
                                      comparator="", value=True)

                self.petri_net_logic.set_awaited_events(task, [to_done_event])

                self.start_tos(task, transport_order.pickup_tos)

            for callback in self.materialflow_callbacks.next_to_cb:
                callback(self._uuid, transport_orders)

    def start_tos(self, task, tos, pickup=True):
        """ Starts scheduling of the given TransportOrderStep """
        if tos.triggered_by:
            self.petri_net_logic.set_awaited_events(task, tos.triggered_by)

            if pickup:
                task.transport_order.state = TransportOrder.TransportOrderState.PICKUP_WAIT_FOR_TRIGGERED_BY
            else:
                task.transport_order.state = TransportOrder.TransportOrderState.DELIVERY_WAIT_FOR_TRIGGERED_BY
            
            uid = self.ids[task.name]
            self.log_transport_order(uid, task.transport_order)
        else:
            tos_started_event = Event(PetriNetConstants.TOS_STARTED_PLACE, "", "Boolean", value=True)
            tos_done_event = Event(PetriNetConstants.TOS_MOVED_TO_LOCATION_PLACE, "", "Boolean", True)
            self.petri_net_logic.set_awaited_events(task, [tos_started_event, tos_done_event])
            self.petri_net_logic.fire_event(task, tos_started_event)

    def on_tos_tb_passed(self, task):
        """
            Gets called when a TriggeredBy is passed in a TransportOrderStep net.
            Set the petr net state and set new awaited event "moved_to_location" for
            either the Pickup Net or the Delivery net depending on the current state
        """
        current_state = task.transport_order.state
        uid = self.ids[task.name]
        transport_order = task.transport_order

        # check the current state and set the new one
        if current_state == TransportOrder.TransportOrderState.PICKUP_WAIT_FOR_TRIGGERED_BY:
            task.transport_order.state = TransportOrder.TransportOrderState.PICKUP_STARTED
        elif current_state == TransportOrder.TransportOrderState.DELIVERY_WAIT_FOR_TRIGGERED_BY:
            task.transport_order.state = TransportOrder.TransportOrderState.DELIVERY_STARTED

        self.log_transport_order(uid, transport_order)

        moved_to_locaction_event = Event("moved_to_location", "", "Boolean", value=True)
        self.petri_net_logic.set_awaited_events(task, [moved_to_locaction_event])

    def on_tos_wait_for_action(self, task):
        """ 
            Gets called when the AGV has moved to the Location.
            Set the petri net state and set FinishedBy events as awaited events.
        """
        current_state = task.transport_order.state
        tos = None
        uid = self.ids[task.name]
        transport_order = task.transport_order

        # check the current state and set the new one
        if current_state == TransportOrder.TransportOrderState.PICKUP_STARTED:
            task.transport_order.state = TransportOrder.TransportOrderState.WAIT_FOR_LOADING
            tos = task.transport_order.pickup_tos
        elif current_state == TransportOrder.TransportOrderState.DELIVERY_STARTED:
            task.transport_order.state = TransportOrder.TransportOrderState.WAIT_FOR_UNLOADING
            tos = task.transport_order.delivery_tos
        else:
            print("Something went wrong!")

        self.log_transport_order(uid, transport_order)

        self.petri_net_logic.set_awaited_events(task, tos.finished_by)

    def on_tos_finished(self, task):
        """
            Gets called when a TransportOrderStep is done.
            Set the petri net state and either call on_pickup_finished method
            or on_delivery_finished method depending on the current state
        """
        current_state = task.transport_order.state
        uid = self.ids[task.name]
        transport_order = task.transport_order

        # check the current state and set the new one
        if current_state == TransportOrder.TransportOrderState.WAIT_FOR_LOADING:
            task.transport_order.state = TransportOrder.TransportOrderState.PICKUP_FINISHED
            self.log_transport_order(uid, transport_order)
            self.on_pickup_finished(task)
        elif current_state == TransportOrder.TransportOrderState.WAIT_FOR_UNLOADING:
            task.transport_order.state = TransportOrder.TransportOrderState.DELIVERY_FINISHED
            self.log_transport_order(uid, transport_order)
            self.on_delivery_finished(task)
        else:
            print("Something went wrong!")

    def on_pickup_finished(self, task):
        """
            Gets called when the Pickup TransportOrderStep is finished.
            Set petri net state and start Delivery TransportOrderStep
        """
        self.pickup_finished(self.ids[task.name])

        task.transport_order.state = TransportOrder.TransportOrderState.DELIVERY_STARTED
        self.start_tos(task, task.transport_order.delivery_tos, False)

    def on_delivery_finished(self, task):
        """
            Gets called when the Delivery TransportOrderStep is finished.
            Set petri net state and fire the to_done event to the task net
        """
        self.delivery_finished(self.ids[task.name])

        to_done_event = Event("to_done", "", "Boolean", value=True) 
        self.petri_net_logic.set_awaited_events(task, [to_done_event])
        self.petri_net_logic.fire_event(task, to_done_event, self.on_petri_net_response)

    def on_to_done(self, task_info):
        """
            Gets called when transport order is done by the AGV.
            Set petri net state (wait for possible FinishedBy events)
        """
        uid = self.ids[task_info.name]
        if self.finished_by_events[task_info.name]:
            transport_order = task_info.transport_order
            transport_order.state = TransportOrder.TransportOrderState.TASK_WAIT_FOR_FINISHED_BY
            self.log_transport_order(uid, transport_order)

            finished_by_events = self.finished_by_events[task_info.name]
            self.petri_net_logic.set_awaited_events(task_info, finished_by_events)
            self.wait_for_finished_by(uid, self.finished_by_events[task_info.name])

    def on_task_finished(self, task_info):
        """
            Gets called when task is finished.
            Starts possible onDone tasks and set petri net state
        """
        uid = self.ids[task_info.name]
        self.task_finished(uid)
        self.tasks_done[task_info.name] = True

        self.petri_net_logic.set_awaited_events(task_info, [None])
        task_info.transport_order.state = TransportOrder.TransportOrderState.FINISHED
        self.log_transport_order(uid, task_info.transport_order)

        if task_info.on_done:
            startable_tasks = []
            for task in task_info.on_done:
                self.not_done_parents[task] = self.not_done_parents[task] - 1

                # all parent tasks are done start the task
                if self.not_done_parents[task] == 0:
                    task_key = self.tasks[str(self.ids[task])]
                    startable_tasks.append(task_key)
                    self.not_done_parents[task_key.name] = self.parent_count[task_key.name]

            self.start_tasks(startable_tasks)
        elif self.all_tasks_done():
            self._is_running = False
            self.all_finished()

    def all_tasks_done(self):
        """
            Returns true if all tasks are done
        """
        if self.cycle_in_call_graph is False:
            for task_done in self.tasks_done.values():
                if task_done is False:
                    return False
            return True
        return False

    def log_transport_order(self, to_uuid, transport_order):
        """
            Saves the given TransportOrder with its locations in the db 
            by calling insert method of the logger
        """
        pickup_location = transport_order.pickup_tos.location
        delivery_location = transport_order.delivery_tos.location
        self.logger.insert_transport_order(self._uuid, to_uuid, transport_order.state,
                                           pickup_location, delivery_location)

    def wait_for_triggered_by(self, uuid_, event_information):
        for callback in self.materialflow_callbacks.triggered_by_cb:
            callback(self._uuid, uuid_, event_information)

    def wait_for_finished_by(self, uuid_, event_information):
        for callback in self.materialflow_callbacks.finished_by_cb:
            callback(self._uuid, uuid_,  event_information)

    def task_finished(self, uuid_):
        for callback in self.materialflow_callbacks.task_finished_cb:
            callback(self._uuid, uuid_)

    def all_finished(self):
        for callback in self.materialflow_callbacks.all_finished_cb:
            callback(self._uuid)

    def pickup_finished(self, uuid_):
        for callback in self.materialflow_callbacks.pickup_finished_cb:
            callback(self._uuid, uuid_)

    def delivery_finished(self, uuid_):
        for callback in self.materialflow_callbacks.delivery_finished_cb:
            callback(self._uuid, uuid_)

    def register_callback_triggered_by(self, callback):
        """
            If a Task can be started and has a TriggeredBy defined, all
            registered callback functions will be called
        """
        if callback not in self.materialflow_callbacks.triggered_by_cb:
            self.materialflow_callbacks.triggered_by_cb.append(callback)

    def register_callback_next_to(self, callback):
        """
            If a Task was started and the TriggeredBy condition is satisfied or there is
            no TriggeredBy all callback functions registered here will be called
        """
        if callback not in self.materialflow_callbacks.next_to_cb:
            self.materialflow_callbacks.next_to_cb.append(callback)

    def register_callback_finished_by(self, callback):
        """
            Functions passed in to this method will be called when the TransportOrder is done
            which means a "to_done" event was sent and a FinishedBy was defined
        """
        if callback not in self.materialflow_callbacks.finished_by_cb:
            self.materialflow_callbacks.finished_by_cb.append(callback)

    def register_callback_task_finished(self, callback):
        """
            If a Task is finished functions registered here are being called.
        """
        if callback not in self.materialflow_callbacks.task_finished_cb:
            self.materialflow_callbacks.task_finished_cb.append(callback)

    def register_callback_all_finished(self, callback):
        """
            If all Tasks in a Materialflow are finished functions registered here are being called
        """
        if callback not in self.materialflow_callbacks.all_finished_cb:
            self.materialflow_callbacks.all_finished_cb.append(callback)

    def register_callback_pickup_finished(self, callback):
        """
            Functions passed in to this method will be called when the Pickup TransportOrderStep
            of a task is finished
        """
        if callback not in self.materialflow_callbacks.pickup_finished_cb:
            self.materialflow_callbacks.pickup_finished_cb.append(callback)

    def register_callback_delivery_finished(self, callback):
        """
            Functions passed in to this method will be called when the Delivery TransportOrderStep
            of a task is finished
        """
        if callback not in self.materialflow_callbacks.delivery_finished_cb:
            self.materialflow_callbacks.delivery_finished_cb.append(callback)



""" Contains Template class """

class Template(object):
    '''
        Represents a Template and all its data
    '''

    def __init__(self):
        self.name = None
        self.keyval = []
        self.context = None

    def __repr__(self):
        return "{}".format(self.keyval)



# Generated from LoTLanParser.g4 by ANTLR 4.8
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .LoTLanParser import LoTLanParser
else:
    from LoTLanParser import LoTLanParser

# This class defines a complete listener for a parse tree produced by LoTLanParser.
class LoTLanParserListener(ParseTreeListener):

    # Enter a parse tree produced by LoTLanParser#program.
    def enterProgram(self, ctx:LoTLanParser.ProgramContext):
        pass

    # Exit a parse tree produced by LoTLanParser#program.
    def exitProgram(self, ctx:LoTLanParser.ProgramContext):
        pass


    # Enter a parse tree produced by LoTLanParser#template.
    def enterTemplate(self, ctx:LoTLanParser.TemplateContext):
        pass

    # Exit a parse tree produced by LoTLanParser#template.
    def exitTemplate(self, ctx:LoTLanParser.TemplateContext):
        pass


    # Enter a parse tree produced by LoTLanParser#templateStart.
    def enterTemplateStart(self, ctx:LoTLanParser.TemplateStartContext):
        pass

    # Exit a parse tree produced by LoTLanParser#templateStart.
    def exitTemplateStart(self, ctx:LoTLanParser.TemplateStartContext):
        pass


    # Enter a parse tree produced by LoTLanParser#instance.
    def enterInstance(self, ctx:LoTLanParser.InstanceContext):
        pass

    # Exit a parse tree produced by LoTLanParser#instance.
    def exitInstance(self, ctx:LoTLanParser.InstanceContext):
        pass


    # Enter a parse tree produced by LoTLanParser#instanceStart.
    def enterInstanceStart(self, ctx:LoTLanParser.InstanceStartContext):
        pass

    # Exit a parse tree produced by LoTLanParser#instanceStart.
    def exitInstanceStart(self, ctx:LoTLanParser.InstanceStartContext):
        pass


    # Enter a parse tree produced by LoTLanParser#memberVariable.
    def enterMemberVariable(self, ctx:LoTLanParser.MemberVariableContext):
        pass

    # Exit a parse tree produced by LoTLanParser#memberVariable.
    def exitMemberVariable(self, ctx:LoTLanParser.MemberVariableContext):
        pass


    # Enter a parse tree produced by LoTLanParser#value.
    def enterValue(self, ctx:LoTLanParser.ValueContext):
        pass

    # Exit a parse tree produced by LoTLanParser#value.
    def exitValue(self, ctx:LoTLanParser.ValueContext):
        pass


    # Enter a parse tree produced by LoTLanParser#transportOrderStep.
    def enterTransportOrderStep(self, ctx:LoTLanParser.TransportOrderStepContext):
        pass

    # Exit a parse tree produced by LoTLanParser#transportOrderStep.
    def exitTransportOrderStep(self, ctx:LoTLanParser.TransportOrderStepContext):
        pass


    # Enter a parse tree produced by LoTLanParser#tosStart.
    def enterTosStart(self, ctx:LoTLanParser.TosStartContext):
        pass

    # Exit a parse tree produced by LoTLanParser#tosStart.
    def exitTosStart(self, ctx:LoTLanParser.TosStartContext):
        pass


    # Enter a parse tree produced by LoTLanParser#tosStatement.
    def enterTosStatement(self, ctx:LoTLanParser.TosStatementContext):
        pass

    # Exit a parse tree produced by LoTLanParser#tosStatement.
    def exitTosStatement(self, ctx:LoTLanParser.TosStatementContext):
        pass


    # Enter a parse tree produced by LoTLanParser#locationStatement.
    def enterLocationStatement(self, ctx:LoTLanParser.LocationStatementContext):
        pass

    # Exit a parse tree produced by LoTLanParser#locationStatement.
    def exitLocationStatement(self, ctx:LoTLanParser.LocationStatementContext):
        pass


    # Enter a parse tree produced by LoTLanParser#optTosStatement.
    def enterOptTosStatement(self, ctx:LoTLanParser.OptTosStatementContext):
        pass

    # Exit a parse tree produced by LoTLanParser#optTosStatement.
    def exitOptTosStatement(self, ctx:LoTLanParser.OptTosStatementContext):
        pass


    # Enter a parse tree produced by LoTLanParser#eventStatement.
    def enterEventStatement(self, ctx:LoTLanParser.EventStatementContext):
        pass

    # Exit a parse tree produced by LoTLanParser#eventStatement.
    def exitEventStatement(self, ctx:LoTLanParser.EventStatementContext):
        pass


    # Enter a parse tree produced by LoTLanParser#onDoneStatement.
    def enterOnDoneStatement(self, ctx:LoTLanParser.OnDoneStatementContext):
        pass

    # Exit a parse tree produced by LoTLanParser#onDoneStatement.
    def exitOnDoneStatement(self, ctx:LoTLanParser.OnDoneStatementContext):
        pass


    # Enter a parse tree produced by LoTLanParser#parameterStatement.
    def enterParameterStatement(self, ctx:LoTLanParser.ParameterStatementContext):
        pass

    # Exit a parse tree produced by LoTLanParser#parameterStatement.
    def exitParameterStatement(self, ctx:LoTLanParser.ParameterStatementContext):
        pass


    # Enter a parse tree produced by LoTLanParser#task.
    def enterTask(self, ctx:LoTLanParser.TaskContext):
        pass

    # Exit a parse tree produced by LoTLanParser#task.
    def exitTask(self, ctx:LoTLanParser.TaskContext):
        pass


    # Enter a parse tree produced by LoTLanParser#taskStart.
    def enterTaskStart(self, ctx:LoTLanParser.TaskStartContext):
        pass

    # Exit a parse tree produced by LoTLanParser#taskStart.
    def exitTaskStart(self, ctx:LoTLanParser.TaskStartContext):
        pass


    # Enter a parse tree produced by LoTLanParser#taskStatement.
    def enterTaskStatement(self, ctx:LoTLanParser.TaskStatementContext):
        pass

    # Exit a parse tree produced by LoTLanParser#taskStatement.
    def exitTaskStatement(self, ctx:LoTLanParser.TaskStatementContext):
        pass


    # Enter a parse tree produced by LoTLanParser#constraintsStatement.
    def enterConstraintsStatement(self, ctx:LoTLanParser.ConstraintsStatementContext):
        pass

    # Exit a parse tree produced by LoTLanParser#constraintsStatement.
    def exitConstraintsStatement(self, ctx:LoTLanParser.ConstraintsStatementContext):
        pass


    # Enter a parse tree produced by LoTLanParser#transportOrder.
    def enterTransportOrder(self, ctx:LoTLanParser.TransportOrderContext):
        pass

    # Exit a parse tree produced by LoTLanParser#transportOrder.
    def exitTransportOrder(self, ctx:LoTLanParser.TransportOrderContext):
        pass


    # Enter a parse tree produced by LoTLanParser#fromStatement.
    def enterFromStatement(self, ctx:LoTLanParser.FromStatementContext):
        pass

    # Exit a parse tree produced by LoTLanParser#fromStatement.
    def exitFromStatement(self, ctx:LoTLanParser.FromStatementContext):
        pass


    # Enter a parse tree produced by LoTLanParser#toStatement.
    def enterToStatement(self, ctx:LoTLanParser.ToStatementContext):
        pass

    # Exit a parse tree produced by LoTLanParser#toStatement.
    def exitToStatement(self, ctx:LoTLanParser.ToStatementContext):
        pass


    # Enter a parse tree produced by LoTLanParser#parameters.
    def enterParameters(self, ctx:LoTLanParser.ParametersContext):
        pass

    # Exit a parse tree produced by LoTLanParser#parameters.
    def exitParameters(self, ctx:LoTLanParser.ParametersContext):
        pass


    # Enter a parse tree produced by LoTLanParser#repeatStatement.
    def enterRepeatStatement(self, ctx:LoTLanParser.RepeatStatementContext):
        pass

    # Exit a parse tree produced by LoTLanParser#repeatStatement.
    def exitRepeatStatement(self, ctx:LoTLanParser.RepeatStatementContext):
        pass


    # Enter a parse tree produced by LoTLanParser#expression.
    def enterExpression(self, ctx:LoTLanParser.ExpressionContext):
        pass

    # Exit a parse tree produced by LoTLanParser#expression.
    def exitExpression(self, ctx:LoTLanParser.ExpressionContext):
        pass


    # Enter a parse tree produced by LoTLanParser#binOperation.
    def enterBinOperation(self, ctx:LoTLanParser.BinOperationContext):
        pass

    # Exit a parse tree produced by LoTLanParser#binOperation.
    def exitBinOperation(self, ctx:LoTLanParser.BinOperationContext):
        pass


    # Enter a parse tree produced by LoTLanParser#unOperation.
    def enterUnOperation(self, ctx:LoTLanParser.UnOperationContext):
        pass

    # Exit a parse tree produced by LoTLanParser#unOperation.
    def exitUnOperation(self, ctx:LoTLanParser.UnOperationContext):
        pass


    # Enter a parse tree produced by LoTLanParser#con.
    def enterCon(self, ctx:LoTLanParser.ConContext):
        pass

    # Exit a parse tree produced by LoTLanParser#con.
    def exitCon(self, ctx:LoTLanParser.ConContext):
        pass



del LoTLanParser


""" Contains Event class """

class Event:
    """ Represents a LoTLan Event instance """

    def __init__(self, logical_name, physical_name, event_type, comparator=None, value=None):
        self.logical_name = logical_name
        self.physical_name = physical_name
        self.event_type = event_type
        self.comparator = comparator
        self.value = value

    def __repr__(self):
        return (("Logical_Name: {}, Physical_Name: {}, Event_Type: {}, Comparator: {}, Value: {}")
                .format(self.logical_name, self.physical_name, self.event_type,
                        self.comparator, self.value))



""" Contains the PetriNetGenerator class """

# 3rd party packages
import snakes.plugins as plugins

# local sources
import lotlan_scheduler.helpers as helpers

from lotlan_scheduler.petri_net.drawer import PetriNetDrawer
from lotlan_scheduler.api.transportorder import TransportOrder
from lotlan_scheduler.api.event import Event

# global defines
from lotlan_scheduler.defines import PetriNetConstants, LogicConstants

plugins.load(["labels", "gv"], "snakes.nets", "nets")
from nets import PetriNet, Place, Transition, Inhibitor, Value, Marking

class TaskRepresentation:
    """
        A representation of a LoTLan Task which contains
        all nessessary data for scheduling
    """
    def __init__(self):
        self.net = None
        self.pickup_net = None
        self.delivery_net = None
        self.awaited_events = []
        self.transition_fired = {}
        self.event_dict = {}

class PetriNetGenerator:
    """
        Takes a lotlan file as argument and generates a Petri Net
        which visualizes the connection between tasks and transportordersteps.
        For each task there is a png file generated
    """

    def __init__(self, tasks, event_instances,
                 test_flag=False, simple_representation=False):
        self.tasks = tasks
        self.petri_net_drawer = PetriNetDrawer()
        self.simple_representation = simple_representation
        self.test_flag = test_flag
        self.event_counter = 0
        self.net_task_dict = {}
        self.event_instances = event_instances
        self.task_representations = {}

    def generate_task_nets(self):
        """ Generates a petri net for each task in tasks """

        self.task_representations = {}
        for task in self.tasks:
            task_repr = TaskRepresentation()
            self.task_representations[task.name] = task_repr
            task_net, pickup_net, delivery_net = self.generate_task_net(task)
            task_repr.net = task_net
            task_repr.pickup_net = pickup_net
            task_repr.delivery_net = delivery_net

            if self.test_flag:
                self.draw_petri_net(task.name, task_net)
                self.draw_petri_net(task.name + "_pickup", pickup_net)
                self.draw_petri_net(task.name + "_delivery", delivery_net)

            for transition in task_net._trans:
                task_repr.transition_fired[transition] = False
            for transition in pickup_net._trans:
                task_repr.transition_fired[transition] = False
            for transition in delivery_net._trans:
                task_repr.transition_fired[transition] = False

        return self.task_representations

    def generate_task_net(self, task):
        """
            Generates a petri net for the task and the TransportOrderSteps of the TransportOrders
            Returns generated nets (Task, Pickup, Delivery)
        """
        task_net = PetriNet(task.name)
        self.net_task_dict[task.name] = task_net

        # generate task action
        self.generate_task_operation(task.name, task_net)

        # generate places and transitions for TriggeredBy
        task.triggered_by_events = self.generate_triggered_by(task, task_net)

        # generate places and transitions for FinishedBy
        task.finished_by_events = self.generate_finished_by(task, task_net)

        pickup_tos = task.transport_order.pickup_tos
        delivery_tos = task.transport_order.delivery_tos

        pickup_net = self.generate_tos_net(task, pickup_tos, task.name + "_pickup")
        delivery_net = self.generate_tos_net(task, delivery_tos, task.name + "_delivery")

        return (task_net, pickup_net, delivery_net)

    def generate_tos_net(self, task, tos, name):
        """
            Generates a petri net for the given TransportOrderStep
        """
        tos_net = PetriNet(name)
        self.generate_advanced_tos_operation(tos_net)
        self.generate_triggered_by_for_tos(task, tos, tos_net)
        self.generate_finished_by_for_tos(task, tos, tos_net)
        return tos_net

    def generate_task_operation(self, task_name, net):
        if self.simple_representation is True:
            self.generate_simple_task_operation(task_name, net)
        else:
            self.generate_advanced_task_operation(net)

    def generate_simple_task_operation(self, task_name, net):
        create_place(task_name, "action", net)

    def generate_advanced_task_operation(self, net):
        """
            Generates a advanced representation for a Task
            contrary to the simple representation this one has
            more places for each state in a Task
        """
        create_place(PetriNetConstants.TASK_STARTED_PLACE,
                     PetriNetConstants.TASK_STARTED_PLACE, net)
        create_place(PetriNetConstants.TASK_DONE_PLACE, PetriNetConstants.TASK_DONE_PLACE, net)
        create_place("to_done", "to_done", net)
        create_place("task_finished", "task_finished", net)

        create_transition(PetriNetConstants.TASK_FIRST_TRANSITION,
                               PetriNetConstants.TASK_FIRST_TRANSITION, net)
        create_transition(PetriNetConstants.TASK_SECOND_TRANSITION,
                               PetriNetConstants.TASK_SECOND_TRANSITION, net)

        net.add_input(PetriNetConstants.TASK_STARTED_PLACE,
                      PetriNetConstants.TASK_FIRST_TRANSITION, Value(1))
        net.add_output(PetriNetConstants.TASK_DONE_PLACE,
                       PetriNetConstants.TASK_FIRST_TRANSITION, Value(1))
        net.add_input("to_done", PetriNetConstants.TASK_FIRST_TRANSITION, Value(1))
        net.add_input(PetriNetConstants.TASK_DONE_PLACE,
                      PetriNetConstants.TASK_SECOND_TRANSITION, Value(1))
        net.add_output("task_finished", PetriNetConstants.TASK_SECOND_TRANSITION, Value(1))

    def generate_advanced_tos_operation(self, net):
        """
            Generates a advanced representation for a TransportOrderStep
            contrary to the simple representation this one has
            more places for each state in a Task
        """
        create_place(PetriNetConstants.TOS_STARTED_PLACE,
                     PetriNetConstants.TOS_STARTED_PLACE, net)
        create_place(PetriNetConstants.TOS_MOVED_TO_LOCATION_PLACE,
                     PetriNetConstants.TOS_MOVED_TO_LOCATION_PLACE, net)
        create_place(PetriNetConstants.TOS_WAIT_FOR_ACTION_PLACE,
                     PetriNetConstants.TOS_WAIT_FOR_ACTION_PLACE, net)
        create_place(PetriNetConstants.TOS_FINISHED_PLACE,
                     PetriNetConstants.TOS_FINISHED_PLACE, net)

        create_transition(PetriNetConstants.TOS_FIRST_TRANSITION,
                          PetriNetConstants.TOS_FIRST_TRANSITION, net)
        create_transition(PetriNetConstants.TOS_SECOND_TRANSITION,
                          PetriNetConstants.TOS_SECOND_TRANSITION, net)

        net.add_input(PetriNetConstants.TOS_STARTED_PLACE,
                      PetriNetConstants.TOS_FIRST_TRANSITION, Value(1))
        net.add_output(PetriNetConstants.TOS_WAIT_FOR_ACTION_PLACE,
                       PetriNetConstants.TOS_FIRST_TRANSITION, Value(1))
        net.add_input(PetriNetConstants.TOS_MOVED_TO_LOCATION_PLACE,
                      PetriNetConstants.TOS_FIRST_TRANSITION, Value(1))
        net.add_input(PetriNetConstants.TOS_WAIT_FOR_ACTION_PLACE,
                      PetriNetConstants.TOS_SECOND_TRANSITION, Value(1))
        net.add_output(PetriNetConstants.TOS_FINISHED_PLACE,
                       PetriNetConstants.TOS_SECOND_TRANSITION, Value(1))

    def draw_petri_net(self, name, petri_net):
        """ Saves the given petri net as an image in the current working directory """
        file_path = "./" + name + PetriNetConstants.IMAGE_ENDING
        self.petri_net_drawer.draw_image(petri_net, file_path)

    def generate_triggered_by(self, task, net):
        """
            Generates places and transitions for the TriggeredBy expression
            of the given Task
        """
        triggered_by_event_names = []
        if task.triggered_by:
            create_transition(PetriNetConstants.TRIGGERED_BY_TRANSITION, "", net)

            if self.simple_representation is True:
                create_place(PetriNetConstants.TRIGGERED_BY, "", net)
                net.add_input(PetriNetConstants.TRIGGERED_BY,
                              PetriNetConstants.TRIGGERED_BY_TRANSITION, Value(1))
                net.add_output(PetriNetConstants.TASK_STARTED_PLACE,
                               PetriNetConstants.TRIGGERED_BY_TRANSITION, Value(1))
            else:
                self.generate_places_from_expression(task.triggered_by, task.name,
                                      PetriNetConstants.TRIGGERED_BY_TRANSITION,
                                      triggered_by_event_names,
                                      False,
                                      tb_event = True)
                net.add_output(PetriNetConstants.TASK_STARTED_PLACE,
                               PetriNetConstants.TRIGGERED_BY_TRANSITION, Value(1))
        return triggered_by_event_names

    def generate_finished_by(self, task, net):
        """
            Generates places and transitions for the FinishedBy expression
            of the given Task
        """
        finished_by_event_names = []
        if self.simple_representation is True:
            create_place(PetriNetConstants.FINISHED_BY, "", net)
            net.add_input(PetriNetConstants.FINISHED_BY, PetriNetConstants.ON_DONE_ENDING,
                          Value(1))
        else:
            if task.finished_by != "":
                self.generate_places_from_expression(task.finished_by, task.name,
                                      PetriNetConstants.TASK_SECOND_TRANSITION,
                                      finished_by_event_names, False, tb_event=False)
        return finished_by_event_names

    def generate_triggered_by_for_tos(self, task, tos, net):
        """
            Generates places and transitions for the TriggeredBy expression
            of the given TransportOrderStep
        """
        tb_statement = tos.triggered_by_statements
        if tb_statement:
            create_transition(PetriNetConstants.TOS_TRIGGERED_BY_TRANSITION, "", net)
            self.generate_places_from_expression(tb_statement, task.name,
                                                PetriNetConstants.TOS_TRIGGERED_BY_TRANSITION,
                                                [],
                                                False,
                                                tb_event = True,
                                                tos_net=net)
            net.add_output(PetriNetConstants.TOS_STARTED_PLACE,
                        PetriNetConstants.TOS_TRIGGERED_BY_TRANSITION, Value(1))

    def generate_finished_by_for_tos(self, task, tos, net):
        """
            Generates places and transitions for the FinishedBy expression
            of the given TransportOrderStep
        """
        fb_statement = tos.finished_by_statements
        if fb_statement:
            self.generate_places_from_expression(fb_statement, task.name,
                                                PetriNetConstants.TOS_SECOND_TRANSITION,
                                                [],
                                                False,
                                                tb_event = False,
                                                tos_net=net)

    def generate_places_from_expression(self, expression, task_name, parent_transition, event_name_list, parent_is_not, tb_event, comparator="", value=None, tos_net=None):
        """ Extracts the single events from the expression and creates a place for each """

        net = None
        if tos_net is not None:
            net = tos_net
        else:
            net = self.net_task_dict[task_name]

        if isinstance(expression, str):
            # generate single node
            event_postfix = ""

            uuid_str = expression + "_" + str(self.event_counter)

            self.event_counter = self.event_counter + 1
            place_name = uuid_str + event_postfix

            event_type = self.event_instances[expression].keyval["type"]
            event = Event(expression, "", event_type, comparator=comparator, value=value)
            create_place(place_name, "", net, event=event, initialized=False)
            net.place(place_name).label(tb_event=tb_event)

            if expression not in self.task_representations[task_name].event_dict:
                self.task_representations[task_name].event_dict[expression] = []
    
            self.task_representations[task_name].event_dict[expression].append(uuid_str)

            event_name_list.append(expression)
            if parent_is_not is True:
                net.add_input(place_name, parent_transition, Inhibitor(Value(1)))
            else:
                net.add_input(place_name, parent_transition, Value(1))

            return expression + event_postfix

        elif isinstance(expression, dict):
            if len(expression) == 2:
                new_parent_is_not = parent_is_not is not True
                self.generate_places_from_expression(expression["value"], task_name,
                                                     parent_transition,
                                                     event_name_list,
                                                     new_parent_is_not, tb_event, tos_net=tos_net)
            elif len(expression) == 3:
                if expression["binOp"] == ".":
                    new_expression = str(expression["left"]) + "." + str(expression["right"])
                    self.generate_places_from_expression(new_expression, task_name, net,
                                                         parent_transition,
                                                         event_name_list,
                                                         parent_is_not, tb_event, tos_net=tos_net)
                # case: expr == <True|False>
                elif expression["right"] == "True":
                    if expression["binOp"] == "==":
                        self.generate_places_from_expression(expression["left"],
                                                             task_name, parent_transition,
                                                             event_name_list, False, tb_event, tos_net=tos_net)
                    elif expression["binOp"] == "!=":
                        self.generate_places_from_expression(expression["left"], task_name,
                                                             parent_transition,
                                                             event_name_list, True, tb_event, tos_net=tos_net)
                elif expression["right"] == "False":
                    if expression["binOp"] == "==":
                        self.generate_places_from_expression(expression["left"], task_name,
                                                             parent_transition,
                                                             event_name_list, True, tb_event, tos_net=tos_net)
                    elif expression["binOp"] == "!=":
                        self.generate_places_from_expression(expression["left"], task_name,
                                                             parent_transition,
                                                             event_name_list, False, tb_event, tos_net=tos_net)
                elif expression["left"] == "(" and expression["right"] == ")":
                    return self.generate_places_from_expression(expression["binOp"], task_name,
                                                                parent_transition,
                                                                event_name_list, parent_is_not,
                                                                tb_event, tos_net=tos_net)
                elif expression["binOp"] == "and" or expression["binOp"] == "or":
                    composition = str(expression)

                    if expression["binOp"] == "and":
                        create_place(composition + "_end", "and", net)
                        create_transition(composition + "_t", "", net)

                        self.generate_places_from_expression(expression["left"], task_name,
                                                             composition + "_t",
                                                             event_name_list, False, tb_event, tos_net=tos_net)
                        self.generate_places_from_expression(expression["right"], task_name,
                                                             composition + "_t",
                                                             event_name_list, False, tb_event, tos_net=tos_net)

                        net.add_output(composition + "_end", composition + "_t", Value(1))

                    elif expression["binOp"] == "or":
                        create_place(composition + "_end", "or", net)

                        create_transition(composition + "_t1", "", net)
                        create_transition(composition + "_t2", "", net)

                        net.add_output(composition + "_end", composition + "_t1", Value(1))
                        net.add_output(composition + "_end", composition + "_t2", Value(1))

                        self.generate_places_from_expression(expression["left"], task_name,
                                                             composition + "_t1",
                                                             event_name_list, False, tb_event, tos_net=tos_net)
                        self.generate_places_from_expression(expression["right"], task_name,
                                                             composition + "_t2",
                                                             event_name_list, False, tb_event, tos_net=tos_net)

                    if parent_is_not is True:
                        net.add_input(composition + "_end", parent_transition, Inhibitor(Value(1)))
                    else:
                        net.add_input(composition + "_end", parent_transition, Value(1))
                elif helpers.is_value(expression["right"]):
                    self.generate_places_from_expression(expression["left"], task_name,
                                                         parent_transition,
                                                         event_name_list, parent_is_not, tb_event,
                                                         expression["binOp"], expression["right"], tos_net=tos_net)

def create_place(place_name, place_type, net, event = None, initialized = True):
    """
        Utility function for creating a place with the snakes module.
        It is used to add a place with the given name and to add labels for
        scheduling (for example if the place represents an event or if its initialized)
    """
    if net.has_place(place_name) is False:
        net.add_place(Place(place_name, []))
        net.place(place_name).label(placeType=place_type, event=event, initialized=initialized)

def create_transition(transition_name, transition_type, net):
    """
        Utility function for creating a transition with the snakes module.
        It is used to add a transition with the given name and to add labels for
        scheduling (currently only the type of the transition)
    """
    net.add_transition(Transition(transition_name))
    net.transition(transition_name).label(transitionType=transition_type)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup( 
    name='Lotlan_Scheduler',
    version='1.1.2',
    author="Peter Detzner, Maximilian Hörstrup", 
    description="Scheduler for tasks generated from Lotlan files",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'antlr4-python3-runtime',
        'tabulate',
        'snakes',
        'networkx',
        'matplotlib',
        'sqlalchemy'
    ]
)


""" Functions defined here set attributes for drawing the petri net """

# standard libraries
import threading

# 3rd party lib
import snakes.plugins

# globals defines
from lotlan_scheduler.defines import DrawConstants, PetriNetConstants

snakes.plugins.load(["labels", "gv"], "snakes.nets", "nets")

draw_lock = threading.Lock()

class PetriNetDrawer:
    """
        Draws image for given petri net.
        Dot attributes for drawing are set here.
    """

    def __init__(self):
        pass

    # dot config methods
    def draw_graph(self, graph, attr):
        attr["nodesep"] = DrawConstants.NODE_SEP_VALUE

    def draw_place(self, place, attr):
        """ Set attributes for drawing places """
        if place.label("placeType") == "input":
            attr["xlabel"] = DrawConstants.INPUT_LABEL
        elif place.label("placeType") == "output":
            attr["xlabel"] = DrawConstants.OUTPUT_LABEL
        elif place.label("placeType") == "connector":
            attr["xlabel"] = DrawConstants.CONNECTOR_LABEL
        elif place.label("placeType") == "or":
            attr["xlabel"] = DrawConstants.OR_LABEL
        elif place.label("placeType") == "and":
            attr["xlabel"] = DrawConstants.AND_LABEL
        elif place.label("placeType") is not None and place.label("placeType") != "":
            attr["xlabel"] = place.label("placeType")
        else:
            attr["xlabel"] = place.name

        if 1 in place:
            attr["label"] = "&bull;"
        else:
            attr["label"] = DrawConstants.PLACE_LABEL
        attr["shape"] = DrawConstants.PLACE_SHAPE

    def draw_transition(self, trans, attr):
        """ Set attributes for drawing transitions """
        if trans.label("transitionType") == "t_s":
            attr["xlabel"] = PetriNetConstants.TASK_FIRST_TRANSITION
        elif trans.label("transitionType") == "t_e":
            attr["xlabel"] = PetriNetConstants.TASK_SECOND_TRANSITION
        elif trans.label("transitionType") == "onDone":
            attr["xlabel"] = DrawConstants.TRANSITION_ON_DONE_LABEL
        elif trans.label("transitionType") == "join":
            attr["xlabel"] = DrawConstants.TRANSITION_JOIN_LABEL

        attr["label"] = DrawConstants.TRANSITION_LABEL
        attr["shape"] = DrawConstants.TRANSITION_SHAPE
        attr["height"] = DrawConstants.TRANSITION_HEIGHT
        attr["width"] = DrawConstants.TRANSITION_WIDTH
        attr["fillcolor"] = DrawConstants.TRANSITION_FILL_COLOR

    def draw_arcs(self, arc, attr):
        if isinstance(arc, snakes.nets.Inhibitor):
            attr["arrowhead"] = DrawConstants.INHIBITOR_ARC_ARROW_HEAD
        attr["label"] = ""

    def draw_image(self, net, filename):
        with draw_lock:
            net.draw(filename, DrawConstants.LAYOUT_METHOD, graph_attr=self.draw_graph,
                     arc_attr=self.draw_arcs, place_attr=self.draw_place,
                     trans_attr=self.draw_transition)



"""
    Contains functions to test given lotlan files and strings
    as well as a test enviroment for validating lotlan grammar and semantic checks
"""
# standard libraries
import sys

from contextlib import contextmanager
import codecs
import errno
import argparse

# 3rd party lib
from antlr4 import InputStream, CommonTokenStream
from tabulate import tabulate

sys.path.append("./")

# local sources
import lotlan_scheduler.helpers as helpers
from lotlan_scheduler.parser.lotlan_tree_visitor import LotlanTreeVisitor
from lotlan_scheduler.validation.semantic_validator import SemanticValidator
from lotlan_scheduler.validation.syntax_error_listener import SyntaxErrorListener
from lotlan_scheduler.parser.LoTLanLexer import LoTLanLexer
from lotlan_scheduler.parser.LoTLanParser import LoTLanParser

# globals defines
from lotlan_scheduler.defines import (TEST_FOLDER_PATH, LOG_PATH, TEMPLATE_STRING)

class ErrorInformation():
    '''
        Class for keeping track of syntax and semantic error count
        in a lotlan file
    '''
    def __init__(self, syntax_error_count, semantic_error_count):
        self.syntax_error_count = syntax_error_count
        self.semantic_error_count = semantic_error_count


# redirects the output(stdout) to a given file
@contextmanager
def stdout_redirection(fileobj):
    old = sys.stdout
    sys.stdout = fileobj
    try:
        yield fileobj
    finally:
        sys.stdout = old


# method for testing: if a valid file got an error or vice versa return 0
def test_all_files():
    valid_file_names = helpers.get_lotlan_file_names(TEST_FOLDER_PATH + "Valid/")
    invalid_file_names = helpers.get_lotlan_file_names(TEST_FOLDER_PATH + "Invalid/")

    failed_tests = 0
    test_results = []

    print("Error messages of tests: \n")
    count = 1
    for valid_file in valid_file_names:
        error_information = test_file(valid_file, TEMPLATE_STRING, False)[1]
        syntax_errors = error_information.syntax_error_count
        semantic_errors = error_information.semantic_error_count
        error_count = syntax_errors + semantic_errors
        if error_count != 0:
            test_results.append((count, valid_file, "0 , 0", str(syntax_errors) + " , " + str(semantic_errors),
                                False))
            failed_tests = failed_tests + 1
        else:
            test_results.append((count, valid_file, "0 , 0", "0 , 0", True))
        count = count + 1

    test_results.append(("------", "-----------------------------------------",
                         "-----", "-----", "-----"))

    for invalid_file in invalid_file_names:
        error_information = test_file(invalid_file, TEMPLATE_STRING, False)[1]
        syntax_errors = error_information.syntax_error_count
        semantic_errors = error_information.semantic_error_count

        expected_errors_string = ""

        if "Semantic" in invalid_file:
            expected_errors_string = "0 , 1"
            if syntax_errors == 0 and semantic_errors == 1:
                test_results.append((count, invalid_file, expected_errors_string, "0 , 1", True))
            else:
                test_results.append((count, invalid_file, expected_errors_string,
                                    str(syntax_errors) + " , " + str(semantic_errors), False))
                failed_tests = failed_tests + 1
        else:
            expected_errors_string = ">0 , 0"
            if syntax_errors > 0:
                test_results.append((count, invalid_file, expected_errors_string, str(syntax_errors) + " , 0", True))
            else:
                test_results.append((count, invalid_file, expected_errors_string, "0 , " + str(semantic_errors), False))
                failed_tests = failed_tests + 1

        print("\n")
        count = count + 1

    test_count = len(valid_file_names) + len(invalid_file_names)
    print(tabulate(test_results, headers=["Test Nr.", "Test Name", "Expected error count (syntax , semantic)",
                                          "Error count", "Has passed"], tablefmt="orgtbl"))
    print("Tests passed: {} of {}".format(test_count - failed_tests, test_count))

    # pre-commit script checks if there was errors (checks for return value 1)
    if failed_tests > 0:
        sys.exit(1)


def test_file(file_path, template_path, used_in_extension):
    try:
        lotlan_file = codecs.open(file_path, "r", encoding="utf8")
        return test_string(lotlan_file.read(), template_path, used_in_extension, file_path)
    except IOError as e:
        if e.errno == errno.EACCES:
            print("File exists, but is not readable")
        elif e.errno == errno.ENOENT:
            print("File does not exist")

        raise Exception("Invalid File")


def test_string(string, template_path, used_in_extension, file_path=""):
    lexer = LoTLanLexer(InputStream(string))
    token_stream = CommonTokenStream(lexer)

    parser = LoTLanParser(token_stream)

    error_listener = SyntaxErrorListener(file_path, used_in_extension, token_stream)
    error_list = error_listener.error_strings

    lexer.removeErrorListeners()
    parser.removeErrorListeners()

    parser.addErrorListener(error_listener)

    tree = parser.program()

    visitor = LotlanTreeVisitor(error_listener)
    if error_listener.is_valid() is False:
        return (None, ErrorInformation(error_listener.syntax_error_count, 0), error_list)

    t = visitor.visit(tree)

    # check for syntax errors detected in visitor
    if error_listener.is_valid() is False:
        return (None, ErrorInformation(error_listener.syntax_error_count, 0), error_list)

    # no syntax errors check for semantic now
    templates = load_templates(template_path)
    semantic_validator = SemanticValidator(file_path, t, templates, used_in_extension,
                                            error_listener)
    semantic_validator.validate()

    return (t, ErrorInformation(0, error_listener.semantic_error_count), error_list)


def load_templates(template_string):
    lexer = LoTLanLexer(InputStream(template_string))
    token_stream = CommonTokenStream(lexer)
    parser = LoTLanParser(token_stream)
    tree = parser.program()
    visitor = LotlanTreeVisitor(None)
    return visitor.visit(tree).templates


def main():
    parser = argparse.ArgumentParser(prog="",
                                     description="tests lotlan grammar and error detection")
    parser.add_argument("--log",help="create a log file and write output there",
                        action="store_true")
    args = parser.parse_args()

    if args.log:
        print("Test Output is printed to file:", LOG_PATH)
        with open(LOG_PATH, "w") as out:
            with stdout_redirection(out):
                test_all_files()
    else:
        test_all_files()

    # sys exit for pre-commit script
    sys.exit(0)


if __name__ == "__main__":
    main()



""" Contains Task class """

class Task(object):
    '''
        Represents a Task and all its data
    '''

    def __init__(self):
        self.name = None
        self.triggered_by = ""
        self.finished_by = ""
        self.transport_order = None
        self.on_done = []
        self.repeat = ""
        self.constraints = ""
        self.context = None
        self.context_dict = {}
        self.triggered_by_events = []
        self.finished_by_events = []

    def __repr__(self):
        return (("\n TriggeredBy: {} \n FinishedBy: {} \n TransportOrder: \n{}"
                 "\n OnDone: {} \n Repeat: {} \n Constraints: {} \n ")
                .format(self.triggered_by, self.finished_by, self.transport_order,
                self.on_done, self.repeat, self.constraints))



""" This module contains the SemanticValidator class """

# local sources
import lotlan_scheduler.helpers as helpers

# globals defines
from lotlan_scheduler.defines import TRIGGERED_BY_KEY, FINISHED_BY_KEY, ON_DONE_KEY, LOCATION_KEY

class SemanticValidator:
    '''
        Executes static semantic checks for given lotlan tree
    '''

    def __init__(self, file_path, tree, templates, used_in_extension, error_listener):
        self.file_path = file_path
        self.templates = templates
        self.tree = tree
        self.used_in_extension = used_in_extension
        self.error_listener = error_listener

    def validate(self):
        self.check_instances()
        self.check_transport_order_steps()
        self.check_tasks()

    def check_instances(self):
        """ Executes semantic checks for all instances """
        error_found = False
        for instance in self.tree.instances.values():
            template_name = instance.template_name
            template = helpers.get_template(self.templates, template_name)

            # check if the corresponding template exists
            if template is None:
                msg = ("The Instance '{}' refers to a template" +
                       "that does not exist").format(instance.name)
                context_start = instance.context.start
                self.error_listener.print_error(msg, context_start.line, context_start.column,
                                                len(instance.name), False)
                error_found = True
            # check if the instance variables match with the corresponding template
            else:
                if (not self.check_if_template_attribute_exists(template, instance) or
                    not self.check_if_template_attribute_definied(template, instance)):
                    error_found = True
        return not error_found

    def check_if_template_attribute_exists(self, template, instance):
        """ Check if all attributes in the given instance are definied in the template """
        for instance_attribute in instance.keyval:
            attribute_found = False
            for template_attribute in template.keyval:
                if instance_attribute == template_attribute:
                    attribute_found = True
            if attribute_found is False:
                msg = ("The attribute '{}' in instance '{}' was not defined " +
                       "in the corresponding template").format(instance_attribute, instance.name)

                variable_context = instance.context_dict[instance_attribute]
                self.error_listener.print_error(msg, variable_context.start.line,
                                                variable_context.start.column,
                                                len(instance_attribute), False)
                return False
        return True

    def check_if_template_attribute_definied(self, template, instance):
        """ Check if all attributes definied in the template are set in instance """
        for template_attribute in template.keyval:
            attribute_found = False
            for instance_attribute in instance.keyval:
                if template_attribute == instance_attribute:
                    attribute_found = True
            if attribute_found is False:
                msg = ("The attribute '{}' from the corresponding template was not" +
                       "definied in instance '{}'").format(template_attribute, instance.name)
                self.error_listener.print_error(msg, instance.context.start.line,
                                                instance.context.start.column,
                                                len(template_attribute), False)
                return False
        return True

    def check_transport_order_steps(self):
        """ Executes checks for all TransportOrderSteps """
        error_found = False
        for tos in self.tree.transport_order_steps.values():
            context_dict = tos.context_dict

            if not self.check_locations(tos) or not self.check_on_done(tos):
                error_found = True

            if TRIGGERED_BY_KEY in context_dict:
                if not self.check_expression(tos.triggered_by_statements,
                                             context_dict[TRIGGERED_BY_KEY]):
                    error_found = True
            if FINISHED_BY_KEY in context_dict:
                if not self.check_expression(tos.finished_by_statements,
                                             context_dict[FINISHED_BY_KEY]):
                    error_found = True
        return not error_found

    def check_locations(self, tos):
        """ Executes semantic checks for all location instances """
        context_dict = tos.context_dict

        if LOCATION_KEY in context_dict:
            location = tos.location
            location_instance = helpers.get_instance(self.tree.instances, location.logical_name)
            location_context = context_dict[LOCATION_KEY]

            if location_instance is None:
                msg = ("The location instance '{}' in TransportOrderStep '{}'" +
                      "could not be found").format(location.logical_name, tos.name)
                self.error_listener.print_error(msg, location_context.start.line,
                                                location_context.start.column,
                                                len(location.logical_name), False)
                return False
            elif location_instance.template_name != "Location":
                msg = ("The instance '{}' in TransportOrderStep '{}' is "
                       "not a 'Location' instance but a '{}' instance").format(
                        location.logical_name, tos.name, location_instance.template_name)
                self.error_listener.print_error(msg, location_context.start.line,
                                                location_context.start.column,
                                                len(location.logical_name), False)
                return False

        return True

    def check_tasks(self):
        """ Executes semantic checks for all Tasks """
        error_found = False
        for task in self.tree.tasks.values():
            context_dict = task.context_dict

            if (not self.check_transport_orders(task) or
                not self.check_repeat_or_on_done(task) or
                not self.check_on_done(task)):
                error_found = True
            if TRIGGERED_BY_KEY in task.context_dict:
                if not self.check_expression(task.triggered_by, context_dict[TRIGGERED_BY_KEY]):
                    error_found = True
            if FINISHED_BY_KEY in task.context_dict:
                if self.check_expression(task.finished_by, context_dict[FINISHED_BY_KEY]):
                    error_found = True
        return not error_found

    def check_transport_orders(self, task):
        """ Executes semantic checks for all TransportOrders """
        if task.transport_order is not None:
            # From check
            if self.check_if_tos_is_present(task.transport_order.pickup_tos.name) is False:
                msg = "Task '{}' refers to an unknown TransportOrderStep in 'from': '{}' ".format(
                    task.name, task.transport_order.pickup_tos.name)
                self.error_listener.print_error(msg, task.context.start.line,
                                                task.context.start.column,
                                                len(task.transport_order.pickup_tos.name), False)
                return False

            tos = helpers.get_transport_order_step(self.tree.transport_order_steps,
                                                   task.transport_order.pickup_tos.name)
            if len(task.transport_order.from_parameters) != len(tos.parameters):
                msg = "From has not the same amount of parameters as the transport order step!"
                self.error_listener.print_error(msg, task.context.start.line,
                                                task.context.start.column,
                                                len(task.transport_order.pickup_tos.name), False)
                return False

            # To check
            if self.check_if_tos_is_present(task.transport_order.delivery_tos.name) is False:
                msg = "Task '{}' refers to an unknown TransportOrderStep in 'to' '{}'".format(
                    task.name, task.transport_order.delivery_tos.name)
                self.error_listener.print_error(msg, task.context.start.line,
                                                task.context.start.column,
                                                len(task.transport_order.delivery_tos.name), False)
                return False

            tos = helpers.get_transport_order_step(self.tree.transport_order_steps,
                                                   task.transport_order.delivery_tos.name)
            if len(task.transport_order.to_parameters) != len(tos.parameters):
                msg = "To has not the same amount of parameters as the transport order step!"
                self.error_listener.print_error(msg, task.context.start.line,
                                                task.context.start.column,
                                                len(task.transport_order.delivery_tos.name), False)
                return False
        return True

    def check_on_done(self, task_or_tos):
        """ Checks if the task in on done refers to a known task """
        context_dict = task_or_tos.context_dict
        error_found = False
        for on_done_task in task_or_tos.on_done:
            if self.check_if_task_is_present(on_done_task) is False:
                msg = ("The task name '{}' in the OnDone statement refers " +
                      "to an unknown Task").format(on_done_task)
                self.error_listener.print_error(msg, context_dict[ON_DONE_KEY].start.line,
                                                context_dict[ON_DONE_KEY].start.column,
                                                len(on_done_task), False)
                error_found = True
        return not error_found

    def check_repeat_or_on_done(self, task):
        """ Checks if a task has both repeat and ondone statements """
        if task.repeat and task.on_done:
            msg = ("The task '{}' has both OnDone and Repeat statements."
                   "It is only allowed to have either of them").format(task.name)
            self.error_listener.print_error(msg, task.context.start.line,
                                            task.context.start.column,
                                            len(task.name), False)
            return False
        return True

    def check_expressions(self, expressions, task_or_tos):
        """ Executes checks to test given expressions """
        if len(expressions) > 1:
            msg = ("There is more than one TriggeredBy or FinishedBy " +
                   "Statement in '{}'").format(task_or_tos.name)
            self.error_listener.print_error(msg, task_or_tos.context.start.line,
                                            task_or_tos.context.start.column,
                                            len(task_or_tos.name), False)
            return False
        if (len(expressions) == 1 and
            not self.check_expression(expressions[0].value, expressions[0].context)):
            return False
        return True

    def check_expression(self, expression, context):
        """ Executes checks to test given expression """
        if isinstance(expression, str):
            if not self.check_single_expression(expression, context):
                return False
        elif isinstance(expression, dict):
            if len(expression) == 2:
                if not self.check_unary_operation(expression, context):
                    return False
            else:
                if not self.check_binary_operation(expression, context):
                    return False
        return True

    def check_single_expression(self, expression, context):
        """
            Checks if a single expression is a valid expression
            example for a single expr.: 'buttonPressed'
        """
        if helpers.is_template_instance(self.tree.instances, expression, "Event") is True:
            instance = helpers.get_instance(self.tree.instances, expression)
            if helpers.has_instance_type(instance, "Boolean") is False:
                msg = ("'" + expression +
                       "' has no booelan type so it cant get parsed as single statement")
                self.error_listener.print_error(msg, context.start.line,
                                                context.start.column, 1, False)
                return False
        elif helpers.is_template_instance(self.tree.instances, expression, "Time") is False:
            if expression not in ["True", "true", "False", "false"]:
                msg = "The given expression is not related to a time or event instance"
                self.error_listener.print_error(msg, context.start.line,
                                                context.start.column, 1, False)
                return False
        return True

    def check_unary_operation(self, expression, context):
        """
            Checks if a unary expression is a valid expression
            example for an unary expr.: !buttonPressed
        """
        if self.is_boolean_expression(expression["value"]) is False:
            msg = ("The given expression couldnt be resolved to " +
                   "a boolean so it cant get parsed as a single statement")
            self.error_listener.print_error(msg, context.start.line, context.start.column,
                                            len(expression["value"]), False)
            return False
        return True

    def check_binary_operation(self, expression, context):
        """
            Checks if a binary expression is a valid expression
            example for a binary expr.: 'buttonPressed and !buttonPressed2'
        """
        left = expression["left"]
        right = expression["right"]

        if expression["binOp"] == ".":
            if self.check_if_task_is_present(expression["left"]) is False:
                msg = "The task given in the expression is not defined"
                self.error_listener.print_error(msg, context.start.line,
                                                context.start.column,
                                                len(expression["left"]), False)
                return False
        else:
            if self.is_boolean_expression(right) is False:
                msg = "The right side is not a boolean expression "
                self.error_listener.print_error(msg, context.start.line,
                                                context.start.column, len(right), False)
                return False
            if self.is_boolean_expression(left) is False:
                msg = "The left side is not a boolean expression"
                self.error_listener.print_error(msg, context.start.line,
                                                context.start.column, len(left), False)
                return False
        return True

    def is_boolean_expression(self, expression):
        """ Checks if the given expression can be resolved to a boolean expression """
        if isinstance(expression, str):
            return self.is_condition(expression)
        if isinstance(expression, dict):
            if len(expression) == 2:
                return self.is_boolean_expression(expression["value"])

            # an expression enclosed by parenthesis has the key binOp in the dict
            if expression["left"] == "(" and expression["right"] == ")":
                return self.is_boolean_expression(expression["binOp"])

            return (self.is_boolean_expression(expression["left"]) and
                    self.is_boolean_expression(expression["right"]))
        return False

    def is_condition(self, expression):
        """ Checks if the given expression is a condition """
        return (helpers.is_template_instance(self.tree.instances, expression, "Event")
                or helpers.is_template_instance(self.tree.instances, expression, "Constraint")
                or helpers.string_is_int(expression)
                or helpers.string_is_float(expression)
                or expression in ["True", "true", "False", "false"]
                or str.startswith(expression, '"') and str.endswith(expression, '"'))

    def check_if_tos_is_present(self, instance_name):
        for tos_name in self.tree.transport_order_steps:
            if instance_name == tos_name:
                return True
        return False

    def check_if_task_is_present(self, instance_name):
        for task in self.tree.tasks:
            if instance_name == task:
                return True
        return False



""" Contains LotlanTreeVisitor """

# standard libraries
from enum import Enum
import re

# local sources
from lotlan_scheduler.model.transport.transport import Transport
from lotlan_scheduler.model.transport.template import Template
from lotlan_scheduler.model.transport.instance import Instance
from lotlan_scheduler.model.transport.task import Task

from lotlan_scheduler.api.transportorder_step import TransportOrderStep
from lotlan_scheduler.api.transportorder import TransportOrder
from lotlan_scheduler.api.location import Location
from lotlan_scheduler.api.event import Event

# globals defines
from lotlan_scheduler.defines import (TRIGGERED_BY_KEY, FINISHED_BY_KEY, REPEAT_KEY,
                     ON_DONE_KEY, TRANSPORT_ORDER_KEY, LOCATION_KEY)

from lotlan_scheduler.parser.LoTLanParserVisitor import LoTLanParserVisitor

class OptionaStatement(Enum):
    '''
        Enum to set the type of the optional statement when returned
    '''
    TRIGGERED_BY = 1
    FINISHED_BY = 2
    ON_DONE = 3


class LotlanTreeVisitor(LoTLanParserVisitor):
    '''
        Uses visitor pattern from antlr to traverse the parse tree
        and store program information in a Transport object
    '''

    def __init__(self, error_listener):
        super()
        self.context_object = {}
        self.error_listener = error_listener

    # Visit a parse tree produced by TaskParser#program.
    def visitProgram(self, ctx):
        # Create Program
        self.cp = Transport()
        self.cp.context_object = self.context_object

        if ctx.children:
            for child in ctx.children:
                program_component = self.visit(child)  # Get object Template|Instance|Task|TOS

                # append appropiatly into the corresponding list
                if isinstance(program_component, Template):
                    if program_component.name not in self.cp.templates:
                        self.cp.templates[program_component.name] = (program_component)
                    else:
                        msg = "There is already an template with the same name defined"
                        self.error_listener.print_error(msg, child.start.line, child.start.column, 1)
                if isinstance(program_component, Instance):
                    if program_component.name not in self.cp.instances:
                        self.cp.instances[program_component.name] = (program_component)
                    else:
                        msg = "There is already an instance with the same name defined"
                        self.error_listener.print_error(msg, child.start.line, child.start.column, 1)
                if isinstance(program_component, Task):
                    if program_component.name not in self.cp.tasks:
                        self.cp.tasks[program_component.name] = (program_component)
                    else:
                        msg = "There is already a task with the same name defined"
                        self.error_listener.print_error(msg, child.start.line, child.start.column, 1)
                if isinstance(program_component, TransportOrderStep):
                    if program_component.name not in self.cp.transport_order_steps:
                        self.cp.transport_order_steps[program_component.name] = (program_component)
                    else:
                        msg = "There is already a transport order step with the same name defined"
                        self.error_listener.print_error(msg, child.start.line, child.start.column, 1)

        # Add tos for transport order to task
        for task in self.cp.tasks.values():
            to = task.transport_order
            if to is not None:
                try:
                    to.pickup_tos = self.cp.transport_order_steps[to.pickup_tos.name]
                    to.delivery_tos = self.cp.transport_order_steps[to.delivery_tos.name]
                except KeyError:
                    pass

        for tos in self.cp.transport_order_steps.values():
            event_instances = {}
            for instance in self.cp.instances.values():
                if instance.template_name == "Event":
                    event_instances[instance.name] = instance

            self.get_events_from_tos(tos.triggered_by_statements, event_instances, tos.triggered_by)
            self.get_events_from_tos(tos.finished_by_statements, event_instances, tos.finished_by)

        return self.cp

    def get_events_from_tos(self, expression, events, event_list, value=None, comparator=None):
        if type(expression) == str and expression != "" and expression in events:
            if value is None:
                value = True
            if comparator is None:
                comparator = "=="

            logical_name = expression
            physical_name = events[logical_name].keyval["name"]
            event_type = events[logical_name].keyval["type"]
    
            event_list.append(Event(logical_name, physical_name, event_type, comparator, value))

        elif type(expression) == dict:
            if len(expression) == 2:
                self.get_events_from_tos(expression["value"], events, event_list, value=False, comparator="!")
            elif len(expression) == 3:
                if expression["binOp"] == ".":
                    self.get_events_from_tos(str(expression["left"] + "." + str(expression["right"])), events, event_list)
                elif expression["left"] == "(" and expression["right"] == ")":
                    self.get_events_from_tos(expression["binOp"], events, event_list)
                elif type(expression["right"]) == str:
                    self.get_events_from_tos(str(expression["left"]), events, event_list, value=expression["right"], comparator=expression["binOp"])
                else:
                    self.get_events_from_tos(expression["left"], events, event_list)
                    self.get_events_from_tos(expression["right"], events, event_list)

    # Visit a parse tree produced by TaskParser#template
    def visitTemplate(self, ctx):
        t = Template()
        t.name = self.visitTemplateStart(ctx.templateStart())
        t.context = ctx

        keyval = []
        for child in ctx.memberVariable():
            variable_content = self.visitMemberVariable(child)
            keyval.append(variable_content[0])
        t.keyval = keyval

        return t

    # Visit a parse tree produced by TaskParser#templateStart
    def visitTemplateStart(self, ctx):
        return ctx.TEMPLATE().getText().split(" ")[1]

    # Visit a parse tree produced by TaskParser#instance
    def visitInstance(self, ctx):
        instance = Instance()

        # Retreive Template and Instance name
        names = self.visitInstanceStart(ctx.instanceStart())

        instance.template_name = names[0]
        instance.name = names[1]
        instance.context = ctx

        keyval = {}
        for child in ctx.memberVariable():
            variable_content = self.visitMemberVariable(child)

            variable = variable_content[0]
            value = variable_content[1]
            value = value.replace('"', "")

            keyval[variable] = value

            if variable in instance.context_dict:
                print("multiple definitions of variable " + variable)
            else:
                instance.context_dict[variable] = child
        instance.keyval = keyval

        return instance

    # Visit a parse tree produced by TaskParser#instanceStart
    def visitInstanceStart(self, ctx):
        template_name = ctx.INSTANCE().getText().split(" ")[0]
        instance_name = ctx.STARTS_WITH_LOWER_C_STR().getText()

        return (template_name, instance_name)

    # Visit a parse tree produced by TaskParser#memberVariable.
    def visitMemberVariable(self, ctx):
        assignment_str = ctx.ASSIGNMENT().getText()

        p = re.compile(r'\w+')
        variable_name = p.search(assignment_str).group(0)

        return (variable_name, self.visitValue(ctx.value()))

    # Visit a parse tree produced by TaskParser#value.
    def visitValue(self, ctx):
        value = 0

        if ctx.STRING_VALUE():
            value = ctx.STRING_VALUE().getText()
        elif ctx.NUMERIC_VALUE():
            value = ctx.NUMERIC_VALUE().getText()
        else:
            value = ctx.EMPTY_VALUE().getText()

        return value

    # Visit a parse tree produced by TaskParser#transportOrderStep.
    def visitTransportOrderStep(self, ctx):
        tos = TransportOrderStep()
        tos.name = self.visitTosStart(ctx.tosStart())
        tos.context = ctx

        for child in ctx.tosStatement():
            self.visitTosStatement(child, tos)
        return tos

    # Visit a parse tree produced by TaskParser#tosStart.
    def visitTosStart(self, ctx):
        return ctx.STARTS_WITH_LOWER_C_STR().getText()

    # Visit a parse tree produced by TaskParser#tosStatements.
    def visitTosStatement(self, ctx, tos):
        context_dict = tos.context_dict

        if ctx.optTosStatement():
            values = self.visitOptTosStatement(ctx.optTosStatement())
            context = ctx.optTosStatement()

            if values[1] == OptionaStatement.TRIGGERED_BY:
                if TRIGGERED_BY_KEY not in context_dict:
                    tos.triggered_by_statements = values[0]
                    context_dict[TRIGGERED_BY_KEY] = context
                else:
                    msg = "TriggeredBy is definied multiple times"
                    self.error_listener.print_error(msg, context.start.line, context.start.column, len("TriggeredBy"))
            elif values[1] == OptionaStatement.FINISHED_BY:
                if FINISHED_BY_KEY not in context_dict:
                    tos.finished_by_statements = values[0]
                    context_dict[FINISHED_BY_KEY] = context
                else:
                    msg = "FinishedBy is definied multiple times"
                    self.error_listener.print_error(msg, context.start.line, context.start.column, len("FinishedBy"))
            elif values[1] == OptionaStatement.ON_DONE:
                if ON_DONE_KEY not in context_dict:
                    tos.on_done = values[0]
                    context_dict[ON_DONE_KEY] = context
                else:
                    msg = "OnDone is definied multiple times"
                    self.error_listener.print_error(msg, context.start.line, context.start.column, len("OnDone"))
        elif ctx.locationStatement():
            if LOCATION_KEY not in context_dict:
                tos.location = self.visitLocationStatement(ctx.locationStatement())
                context_dict[LOCATION_KEY] = ctx.locationStatement()
            else:
                context = ctx.locationStatement()
                msg = "Location is definied multiple times"
                self.error_listener.print_error(msg, context.start.line, context.start.column, len("Location"))
        elif ctx.parameterStatement():
            parameters = self.visitParameterStatement(ctx.parameterStatement())
            tos.parameters = parameters

    def visitParameterStatement(self, ctx):
        parameters = []
        for parameter in ctx.STARTS_WITH_LOWER_C_STR():
            parameters.append(parameter.getText())
        return parameters

    # Visit a parse tree produced by TaskParser#Location Statement.
    def visitLocationStatement(self, ctx):
        location = Location(ctx.STARTS_WITH_LOWER_C_STR().getText(), "", "")
        return location

    # Visit a parse tree produced by TaskParser#optTosStatement.
    def visitOptTosStatement(self, ctx):
        childs = ctx.children
        for i in range(len(ctx.children)):
            if childs[i] == ctx.eventStatement():
                return self.visitEventStatement(ctx.eventStatement())
            elif childs[i] == ctx.onDoneStatement():
                return self.visitOnDoneStatement(ctx.onDoneStatement())

    def visitEventStatement(self, ctx):
        if ctx.TRIGGERED_BY():
            return (self.visitExpression(ctx.expression()), OptionaStatement.TRIGGERED_BY)
        elif ctx.FINISHED_BY():
            return (self.visitExpression(ctx.expression()), OptionaStatement.FINISHED_BY)

    def visitOnDoneStatement(self, ctx):
        on_done = []
        for task in ctx.STARTS_WITH_LOWER_C_STR():
            on_done.append(task.getText())
        return (on_done, OptionaStatement.ON_DONE)

    # Visit a parse tree produced by TaskParser.
    def visitTask(self, ctx):
        ti = Task()
        ti.name = self.visitTaskStart(ctx.taskStart())
        ti.context = ctx

        for child in ctx.taskStatement():
            self.visitTaskStatement(child, ti)
        return ti

    # Visit a parse tree produced by TaskParser#taskStart.
    def visitTaskStart(self, ctx):
        return ctx.STARTS_WITH_LOWER_C_STR().getText()

    # Visit a parse tree produced by TaskParser#taskStatement.
    def visitTaskStatement(self, ctx, task_info):
        context_dict = task_info.context_dict

        if(ctx.repeatStatement()):
            repeat_stmt = self.visitRepeatStatement(ctx.repeatStatement())
            if REPEAT_KEY not in context_dict:
                context_dict[REPEAT_KEY] = ctx.repeatStatement()
                task_info.repeat = repeat_stmt
            else:
                context = ctx.repeatStatement()
                msg = "Repeat was defined multiple times!"
                self.error_listener.print_error(msg, context.start.line, context.start.column, len("Repeat"))
        elif ctx.optTosStatement():
            values = self.visitOptTosStatement(ctx.optTosStatement())
            context = ctx.optTosStatement()

            if values[1] == OptionaStatement.TRIGGERED_BY:
                if TRIGGERED_BY_KEY not in context_dict:
                    task_info.triggered_by = values[0]
                    task_info.context_dict[TRIGGERED_BY_KEY] = ctx.optTosStatement()
                else:
                    msg = "TriggeredBy is definied multiple times"
                    self.error_listener.print_error(msg, context.start.line, context.start.column, len("TriggeredBy"))
            elif values[1] == OptionaStatement.FINISHED_BY:
                if FINISHED_BY_KEY not in context_dict:
                    task_info.finished_by = values[0]
                    task_info.context_dict[FINISHED_BY_KEY] = ctx.optTosStatement()
                else:
                    msg = "FinishedBy is definied multiple times"
                    self.error_listener.print_error(msg, context.start.line, context.start.column, len("FinishedBy"))
            elif values[1] == OptionaStatement.ON_DONE:
                if ON_DONE_KEY not in context_dict:
                    task_info.on_done = values[0]
                    context_dict[ON_DONE_KEY] = ctx.optTosStatement()
                else:
                    msg = "OnDone is definied multiple times"
                    self.error_listener.print_error(msg, context.start.line, context.start.column, len("OnDone"))
        elif ctx.transportOrder():
            if TRANSPORT_ORDER_KEY not in context_dict:
                task_info.transport_order = self.visitTransportOrder(ctx.transportOrder())
                context_dict[TRANSPORT_ORDER_KEY] = ctx.optTosStatement()
            else:
                context = ctx.transportOrder()
                msg = "TransportOrder was defined multiple times"
                self.error_listener.print_error(msg, context.start.line, context.start.column, len("Transport"))

    # Visit a parse tree produced by TaskParser#transportOrder.
    def visitTransportOrder(self, ctx):
        transport_order = TransportOrder()

        childs = ctx.children
        for i in range(len(childs)):
            if childs[i] == ctx.fromStatement():
                self.visitFromStatement(ctx.fromStatement(), transport_order)
            elif childs[i] == ctx.toStatement():
                self.visitToStatement(ctx.toStatement(), transport_order)
        return transport_order

    def visitFromStatement(self, ctx, transport_order):
        transport_order.pickup_tos.name = ctx.STARTS_WITH_LOWER_C_STR().getText()
        parameters = ctx.parameters()
        if parameters is not None:
            for parameter in parameters.children:
                if parameter.getText() != ",":
                    transport_order.from_parameters.append(parameter.getText())

    def visitToStatement(self, ctx, transport_order):
        transport_order.delivery_tos.name = ctx.STARTS_WITH_LOWER_C_STR().getText()
        parameters = ctx.parameters()
        if parameters is not None:
            for parameter in parameters.children:
                if parameter.getText() != ",":
                    transport_order.to_parameters.append(parameter.getText())

    def visitParameters(self, ctx):
        parameters = []
        if ctx:
            for parameter in ctx:
                parameters.append(parameter.getText())
        return parameters

    def visitRepeatStatement(self, ctx):
        return ctx.INTEGER().getText()

    # Visit a parse tree produced by TaskParser#expression.
    def visitExpression(self, ctx):
        length = len(ctx.children)

        if length == 1:  # Terminal,  returns TERMINAL
            ele = self._getContent(ctx.children[0])
            return ele
        if length == 2:  # UnOperation, returns {unop: !, value: EXPRESSION}
            unOp = self._getContent(ctx.children[0])
            ele = self._getContent(ctx.children[1])
            return dict(unop=unOp, value=ele)

        if length == 3:  # binOperation:, returns {op: (==|!=|<=|..|), left: EXPRESSION, right:EXPRESSION}
            left = self._getContent(ctx.children[0])
            binOp = self._getContent(ctx.children[1])
            right = self._getContent(ctx.children[2])
            return dict(binOp=binOp, left=left, right=right)

        return None

    def _getContent(self, child):
        ele = self.visit(child)
        # If None, then an instance.value is used!
        if ele is None:
            ele = child.getText()
        return ele

    # Visit a parse tree produced by TaskParser#binOperation.
    def visitBinOperation(self, ctx):
        return ctx.children[0].getText()

    # Visit a parse tree produced by TaskParser#unOperation.
    def visitUnOperation(self, ctx):
        return ctx.children[0].getText()

    # Visit a parse tree produced by TaskParser#con.
    def visitCon(self, ctx):
        return ctx.children[0].getText()



""" Contains SQLLogger class """

# standard libraries
from pathlib import Path
import hashlib
import time
import threading

# 3rd party packages
from sqlalchemy import create_engine, MetaData, Table

# globals defines
from lotlan_scheduler.defines import SQLCommands

init_lock = threading.Lock()
mf_lock = threading.Lock()
to_lock = threading.Lock()
location_lock = threading.Lock()

class SQLiteLogger():
    """
        Establishes a SQLite connection and inserts logging data
    """
    def __init__(self, dabase_path = SQLCommands.DATABASE_PATH):
        with init_lock:
            database_file = Path(dabase_path)
            create_tables = True

            # check if db file is already there
            if database_file.is_file():
                create_tables = False
            self.database_engine = create_engine("sqlite:///" + dabase_path, echo=False)
            self.metadata = MetaData(bind=self.database_engine)
            self.con = self.database_engine.connect()
            self.mf_uuid_to_mf_instance_id = {}

            if create_tables:
                self.database_engine.execute(SQLCommands.CREATE_MATERIALFLOW_TABLE)
                self.database_engine.execute(SQLCommands.CREATE_MATERIALFLOW_INSTANCE_TABLE)
                self.database_engine.execute(SQLCommands.CREATE_TRANSPORT_ORDER_TABLE)
                self.database_engine.execute(SQLCommands.CREATE_TRANSPORT_ORDER_IDS_TABLE)
                self.database_engine.execute(SQLCommands.CREATE_LOCATION_TABLE)

    def insert_materialflow_in_sql(self, mf_uuid, lotlan_string):
        """
            Inserts materialflow information into the tables
            'materialflow' and 'materialflow_instance'
            Returns materailflow_instance id if no error occured
        """

        materialflow_table = Table("materialflow", self.metadata, autoload=True)
        lotlan_hash = hashlib.md5(lotlan_string.encode()).hexdigest()

        mf_uuid = str(mf_uuid)
        materialflow_id = None

        with mf_lock:
            select_stmt = materialflow_table.select(materialflow_table.c.hash==lotlan_hash)
            result = select_stmt.execute().first()

            if result is None:
                result = self.con.execute(materialflow_table.insert(),
                                        lotlan=lotlan_string, hash=lotlan_hash)
                materialflow_id = result.lastrowid
            else:
                materialflow_id = result.id

        if materialflow_id is not None:
            now = int(time.time()) # time in utc
            materialflow_instance_table = Table("materialflow_instance",
                                                self.metadata, autoload=True)
            result = self.con.execute(materialflow_instance_table.insert(),
                                      materialflow_id=materialflow_id,
                                      uuid=str(mf_uuid),
                                      timestamp=now)
            materialflow_instance_id = result.lastrowid

            self.mf_uuid_to_mf_instance_id[mf_uuid] = materialflow_instance_id

    def insert_transport_order(self, mf_uuid, to_uuid, state, pickup, delivery):
        """
            Inserts a TransportOrder state into the database
            pickup and delivery have to be Location objects
        """
        now = int(time.time()) # time in utc
        mf_uuid = str(mf_uuid)

        # Get foreign keys in transport_order table
        transport_uuid = self.get_transport_uuid(to_uuid)
        pickup_id = self.get_location_id(pickup)
        delivery_id = self.get_location_id(delivery)

        mf_instance_id = self.mf_uuid_to_mf_instance_id[mf_uuid]

        transport_order_table = Table("transport_order", self.metadata, autoload=True)
        self.con.execute(transport_order_table.insert(),
                         materialflow_id=mf_instance_id,
                         timestamp=now,
                         transport_uuid=transport_uuid,
                         state=state,
                         location_id_pickup=pickup_id,
                         location_id_delivery =delivery_id)

    def get_transport_uuid(self, to_uuid):
        """
            Searches for foreign key of given TransportOrder.
            If no entry is found a new one is inserted.
        """
        transport_order_ids_table = Table("transport_order_ids", self.metadata, autoload=True)
        transport_uuid = None

        to_uuid = str(to_uuid)

        with to_lock:
            select_stmt = transport_order_ids_table.select(transport_order_ids_table.c.uuid==to_uuid)
            result = select_stmt.execute().first()

            if result is None:
                result = self.con.execute(transport_order_ids_table.insert(), uuid=to_uuid)
                transport_uuid = result.lastrowid
            else:
                transport_uuid = result.id

        return transport_uuid

    def get_location_id(self, location):
        """
            Searches for foreign key for given location
            If no entry is found a new one is inserted.

            Returns the foreign key if no error occured and location object is defined
            Returns None otherwise
        """
        location_id = None
        if location:
            location_table = Table("location", self.metadata, autoload=True)

            with location_lock:
                select_stmt = location_table.select(location_table.c.logical_name==location.logical_name)
                result = select_stmt.execute().first()

                if result is None:
                    insert_result = self.con.execute(location_table.insert(),
                                                    logical_name=location.logical_name,
                                                    physical_name=location.physical_name,
                                                    location_type=location.location_type)
                    location_id = insert_result.lastrowid
                else:
                    location_id = result.id

        return location_id
